
    (*********************************************************************
     *  DCParametricEQ.pas                                               *
     *                                                                   *
     *  This unit is Part of the DC-Param EQ Filter v1.1                 *
     *                                                                   *
     *  author    : Milenko Mitrovic                                     *
     *  email     : dcoder@dsp-worx.de                                   *
     *  web       : http://dsp-worx.de                                   *
     *  date      : 24-07-2003                                           *
     *                                                                   *
     *  The contents of this file are used with permission, subject to   *
     *  the Mozilla Public License Version 1.1 (the "License"); you may  *
     *  not use this file except in compliance with the License. You may *
     *  obtain a copy of the License at                                  *
     *  http://www.mozilla.org/MPL/MPL-1.1.html                          *
     *                                                                   *
     *  Software distributed under the License is distributed on an      *
     *  "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or   *
     *  implied. See the License for the specific language governing     *
     *  rights and limitations under the License.                        *
     *                                                                   *
     *  (C) 2003, 2004 Milenko Mitrovic <dcoder@dsp-worx.de>             *
     *                                                                   *
     *********************************************************************)

unit DCParametricEQ;

{.$DEFINE MEASURE}

interface

uses
  Classes, dspConst, dspUtils, Math, Windows, SysUtils;

type
  TEQBand = record
    F: Single;
    Q: Single;
  end;

const
  EQ_NUM_SPEAKERS = 18;
  EQ_NUM_BANDS = 31;
  EQ_NUM_BANDS_ALIGNED = 32; // must be greater than EQ_NUM_BANDS and multiple of 16

  EQ_BANDS: array[0..EQ_NUM_BANDS-1] of TEQBand = (
    (F: 20.159;    Q: 4.318),
    (F: 25.398;    Q: 4.318),
    (F: 32.0;      Q: 4.318),
    (F: 40.317;    Q: 4.318),
    (F: 50.797;    Q: 4.318),
    (F: 64.0;      Q: 4.318),
    (F: 80.635;    Q: 4.318),
    (F: 101.594;   Q: 4.318),
    (F: 128.0;     Q: 4.318),
    (F: 161.270;   Q: 4.318),
    (F: 203.187;   Q: 4.318),
    (F: 256.0;     Q: 4.318),
    (F: 322.540;   Q: 4.318),
    (F: 406.375;   Q: 4.318),
    (F: 512.0;     Q: 4.318),
    (F: 645.08;    Q: 4.318),
    (F: 812.749;   Q: 4.318),
    (F: 1024.0;    Q: 4.318),
    (F: 1290.159;  Q: 4.318),
    (F: 1625.499;  Q: 4.318),
    (F: 2048.0;    Q: 4.318),
    (F: 2580.318;  Q: 4.318),
    (F: 3250.997;  Q: 4.318),
    (F: 4096.0;    Q: 4.318),
    (F: 5160.637;  Q: 4.318),
    (F: 6501.995;  Q: 4.318),
    (F: 8192.0;    Q: 4.318),
    (F: 10321.273; Q: 4.318),
    (F: 13003.989; Q: 4.318),
    (F: 16384.0;   Q: 4.318),
    (F: 20642.546; Q: 4.318)
  );

type
  TAlignedFloatArray = array[0..31, 0..EQ_NUM_BANDS_ALIGNED -1, 0..7] of Single;
  PAlignedFloatArray = ^TAlignedFloatArray;

  TDCParamEQ = class
  private
    FChannels: integer;
    FChannelMapping: Cardinal;
    FNeedCoeff: Boolean;
    FNeedAmp: Boolean;
    FEnabled : Boolean;
    FSampleRate: integer;
    FAlpha: array[0..EQ_NUM_SPEAKERS -1, 0..EQ_NUM_BANDS-1] of Single;
    FX: PAlignedFloatArray;
    FA: PAlignedFloatArray;

    FXX: array[0..EQ_NUM_SPEAKERS -1, 0..EQ_NUM_BANDS-1, 0..1] of Single;
    FYY: array[0..EQ_NUM_SPEAKERS -1, 0..EQ_NUM_BANDS-1, 0..1] of Single;
    FAA: array[0..EQ_NUM_SPEAKERS -1, 0..EQ_NUM_BANDS-1, 0..1] of Single;
    FBB: array[0..EQ_NUM_SPEAKERS -1, 0..EQ_NUM_BANDS-1, 0..1] of Single;
    FB0: array[0..EQ_NUM_SPEAKERS -1, 0..EQ_NUM_BANDS-1] of Single;

    FGainDB: array[0..EQ_NUM_SPEAKERS -1, 0..EQ_NUM_BANDS-1] of Single;
    FChannelGainDB: array[0..EQ_NUM_SPEAKERS -1, 0..EQ_NUM_BANDS-1] of PSingle;
    FPreAmp: array[0..EQ_NUM_SPEAKERS -1] of Single;
    FChannelPreAmp: array[0..EQ_NUM_SPEAKERS -1] of PSingle;
    procedure SetGainDB(ABand: integer; AChannel: integer; AValue: Single);
    function GetGainDB(ABand: integer; AChannel: integer): Single;
    procedure SetPreAmp(AChannel: integer; AValue: Single);
    function GetPreAmp(AChannel: integer): Single;
    procedure SetEnabled(AEnabled: Boolean);
    function GetEnabled: Boolean;
    function GetChannelMapping: Cardinal;
    procedure CalcCoefficients;
    procedure CalcAmplitude;
    procedure SetupChannels;

    procedure Process8Bit(Buffer: PByte; NumSamples: Integer; Channels: integer);
    procedure Process16Bit(Buffer: PSmallInt; NumSamples: Integer; Channels: integer);
    procedure Process24Bit(Buffer: P24BitSample; NumSamples: Integer; Channels: integer);
    procedure Process32Bit(Buffer: PInteger; NumSamples: Integer; Channels: integer);
    procedure Process32BitFloat(Buffer: PSingle; NumSamples: Integer; Channels: integer);

    procedure Process8BitSSE(Buffer: PByte; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
    procedure Process16BitSSE(Buffer: PSmallInt; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
    procedure Process24BitSSE(Buffer: P24BitSample; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
    procedure Process32BitIntegerSSE(Buffer: PInteger; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
    procedure Process32BitFloatSSE(Buffer: PSingle; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);

    procedure Process8BitSSE3(Buffer: PByte; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
    procedure Process16BitSSE3(Buffer: PSmallInt; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
    procedure Process24BitSSE3(Buffer: P24BitSample; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
    procedure Process32BitIntegerSSE3(Buffer: PInteger; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
    procedure Process32BitFloatSSE3(Buffer: PSingle; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Process(Buffer: Pointer; Size, Samplerate: Integer; Bits: integer; Channels: integer; Float: Boolean; ChannelMapping: Cardinal);
    procedure Flush;
    procedure Reset;
    property GainDB[Band: integer; Channel: integer]: Single read GetGainDB write SetGainDB;
    property PreAmp[Channel: integer]: Single read GetPreAmp write SetPreAmp;
  published
    property Enabled : Boolean read GetEnabled write SetEnabled;
    property ChannelMapping: Cardinal read GetChannelMapping;
    property Channels: Integer read FChannels;
  end;

implementation

const
  LOGFACTOR: Single = 10.0;

  EQ_NUM_BANDS_ALIGNED_SIZE: Integer = EQ_NUM_BANDS_ALIGNED * 8 * SizeOf(Single);
  LOOP_DECREMENT: Integer = 8 * 4 * EQ_NUM_BANDS * -1;

  den: Single = 0.00000001;
  InvertSSE_LL : array[0..3] of Cardinal = (
    0, 0, $80000000, $80000000
  );

  EQ_NUM_BANDS_2 = EQ_NUM_BANDS shl 1;

constructor TDCParamEQ.Create;
var
  i, c: integer;
begin
  inherited Create;

  FEnabled := False;

  FNeedAmp := False;
  FNeedCoeff := True;
  FSampleRate := 0;
  FChannels := 0;
  FChannelMapping := 0;
  if (hasSSE) then
  begin
    FX := AllocMem(SizeOf(TAlignedFloatArray));
    FA := AllocMem(SizeOf(TAlignedFloatArray));
  end;

  for i := 0 to EQ_NUM_BANDS -1 do
    for c := 0 to EQ_NUM_SPEAKERS -1 do
    begin
      FGainDB[c,i] := 0.0;
      FPreAmp[c] := 1.0;
    end;
end;

destructor TDCParamEQ.Destroy;
begin
  if (hasSSE) then
  begin
    FreeMem(FX);
    FreeMem(FA);
  end;

  inherited Destroy;
end;

procedure TDCParamEQ.CalcCoefficients;
var
  omega: Double;
  freq: Double;
  i, c: integer;
begin
  for c := 0 to EQ_NUM_BANDS -1 do
  begin
    if EQ_BANDS[c].F < 1
      then freq := 1
      else if Round(EQ_BANDS[c].F * 2) >= Round(FSampleRate)
        then freq := 0 // FSampleRate / 2
        else freq := EQ_BANDS[c].F;

    for i := 0 to FChannels -1 do
    begin
      omega := (2 * PI * freq) / FSampleRate;
      if hasSSE then
      begin
        FA[i,c,2] := -2 * cos(omega);
        FA[i,c,0] := FA[i,c,2];
      end else
      begin
        FB0[i,c] := -2 * cos(omega);
      end;
      FAlpha[i,c] := sin(omega) / (2.0 * EQ_BANDS[c].Q);
    end;
  end;
  CalcAmplitude;
end;

procedure TDCParamEQ.CalcAmplitude;
var
  i, c: integer;
  amp: Double;
begin
  for c := 0 to EQ_NUM_BANDS -1 do
  begin
    for i := 0 to FChannels -1 do
    begin
      amp := power(10, FChannelGainDB[i,c]^ / (LOGFACTOR * 4));
      if hasSSE then
      begin
        FA[i,c,4] := 1 / (1 + (FAlpha[i,c] / amp));
        FA[i,c,3] := 1 - (FAlpha[i,c] / amp);
        FA[i,c,5] := 1 + (FAlpha[i,c] * amp);
        FA[i,c,1] := 1 - (FAlpha[i,c] * amp);
      end else
      begin
        FBB[i,c,0] := 1 + (FAlpha[i,c] * amp);
        FBB[i,c,1] := 1 - (FAlpha[i,c] * amp);
        FAA[i,c,0] := 1 / (1 + (FAlpha[i,c] / amp));
        FAA[i,c,1] := 1 - (FAlpha[i,c] / amp);
      end;
    end;
  end;
end;

procedure TDCParamEQ.SetupChannels;
var
  i, c: integer;
  p: Cardinal;
  f: Cardinal;
begin
  if (FChannelMapping = 0) then
  begin
    for i := 0 to FChannels -1 do
    begin
      for c := 0 to EQ_NUM_BANDS -1 do
        FChannelGainDB[i,c] := @FGainDB[i,c];
      FChannelPreAmp[i] := @FPreAmp[i];
    end;
    Exit;
  end;

  p := 1;
  i := 0;
  f := 0;

  while (i < FChannels) do
  begin
    if (p and FChannelMapping = p) then
    begin

      for c := 0 to EQ_NUM_BANDS -1 do
        FChannelGainDB[i,c] := @FGainDB[f,c];

      FChannelPreAmp[i] := @FPreAmp[f];

      inc(i);
    end;
    inc(f);

    p := p shl 1;
  end;
end;

procedure TDCParamEQ.Flush;
const
  EQBufferSize = EQ_NUM_SPEAKERS * EQ_NUM_BANDS * SizeOf(Single) * 2;
begin
  if hasSSE then
  begin
    FillChar(FX^, SizeOf(TAlignedFloatArray), 0);
  end else
  begin
    FillChar(FXX, EQBufferSize, 0);
    FillChar(FYY, EQBufferSize, 0);
  end;
end;

procedure TDCParamEQ.SetGainDB(ABand: integer; AChannel: integer; AValue: Single);
begin
  if not InRange(AChannel,0,EQ_NUM_SPEAKERS -1) then Exit;
  if not InRange(ABand,0,EQ_NUM_BANDS -1) then Exit;
  FGainDB[AChannel,ABand] := AValue;
  FNeedAmp := True;
end;

function TDCParamEQ.GetGainDB(ABand: integer; AChannel: integer): Single;
begin
  Result := 0.0;
  if not InRange(AChannel,0,EQ_NUM_SPEAKERS -1) then Exit;
  if not InRange(ABand,0,EQ_NUM_BANDS -1) then Exit;
  Result := FGainDB[AChannel,ABand];
end;

procedure TDCParamEQ.SetPreAmp(AChannel: integer; AValue: Single);
begin
  if not InRange(AChannel,0,EQ_NUM_SPEAKERS -1) then Exit;
  FPreAmp[AChannel] := power(10.0, AValue / LOGFACTOR);
end;

function TDCParamEQ.GetPreAmp(AChannel: integer): Single;
begin
  Result := 0.0;
  if not InRange(AChannel,0,EQ_NUM_SPEAKERS -1) then Exit;
  Result := log10(FPreAmp[AChannel]) * LOGFACTOR;
end;

procedure TDCParamEQ.SetEnabled(AEnabled: Boolean);
begin
  FEnabled := AEnabled;
end;

function TDCParamEQ.GetEnabled: Boolean;
begin
  Result := FEnabled;
end;

procedure TDCParamEQ.Process8Bit(Buffer: PByte; NumSamples: Integer; Channels: integer);
var
  c, n: integer;
  i: integer;
  so, si: Single;
  m1, m2, m3, m4 : PSingle;
  n1: PSingle;
  pa: Single;
begin
  dec(Buffer);
  for c := 0 to Channels -1 do
  begin
    inc(Buffer);
    pa := FChannelPreAmp[c]^;

    m1 := @FXX[c,0,0];
    m2 := @FXX[c,0,1];
    m3 := @FYY[c,0,0];
    m4 := @FYY[c,0,1];

    n1 := @FB0[c,0];

    for i := 0 to NumSamples -1 do
    begin
      si := (Buffer^ - 128) * pa;

      if (si = 0.0)
        then si := 0.00000001;

      for n := 0 to EQ_NUM_BANDS -1 do
      begin

        so := (FBB[c,n,0]*si + n1^*m1^ + FBB[c,n,1]*m2^ - n1^*m3^ - FAA[c,n,1]*m4^) * FAA[c,n,0];

        m2^ := m1^;
        inc(m2, 2);

        m1^ := si;
        inc(m1, 2);

        m4^ := m3^;
        inc(m4, 2);

        m3^ := so;
        inc(m3, 2);

        inc(n1, 1);

        si := so;
      end;

      dec(m1, EQ_NUM_BANDS_2);
      dec(m2, EQ_NUM_BANDS_2);
      dec(m3, EQ_NUM_BANDS_2);
      dec(m4, EQ_NUM_BANDS_2);
      dec(n1, EQ_NUM_BANDS);

      Buffer^ := Clip_8(Round(so)) + 128;
      inc(Buffer, Channels)
    end;

    dec(Buffer, NumSamples*Channels);
  end;
end;

procedure TDCParamEQ.Process16Bit(Buffer: PSmallInt; NumSamples: Integer; Channels: integer);
var
  c, n: integer;
  i: integer;
  so, si: Single;
  m1, m2, m3, m4 : PSingle;
  n1: PSingle;
  pa: Single;
begin
  dec(Buffer);
  for c := 0 to Channels -1 do
  begin
    inc(Buffer);
    pa := FChannelPreAmp[c]^;

    m1 := @FXX[c,0,0];
    m2 := @FXX[c,0,1];
    m3 := @FYY[c,0,0];
    m4 := @FYY[c,0,1];

    n1 := @FB0[c,0];

    for i := 0 to NumSamples -1 do
    begin
      si := Buffer^ * pa;

      if (si = 0.0)
        then si := 0.00000001;

      for n := 0 to EQ_NUM_BANDS -1 do
      begin

        so := (FBB[c,n,0]*si + n1^*m1^ + FBB[c,n,1]*m2^ - n1^*m3^ - FAA[c,n,1]*m4^) * FAA[c,n,0];

        m2^ := m1^;
        inc(m2, 2);

        m1^ := si;
        inc(m1, 2);

        m4^ := m3^;
        inc(m4, 2);

        m3^ := so;
        inc(m3, 2);

        inc(n1, 1);

        si := so;
      end;

      dec(m1, EQ_NUM_BANDS_2);
      dec(m2, EQ_NUM_BANDS_2);
      dec(m3, EQ_NUM_BANDS_2);
      dec(m4, EQ_NUM_BANDS_2);
      dec(n1, EQ_NUM_BANDS);

      Buffer^ := Clip_16(Round(so));
      inc(Buffer, Channels)
    end;

    dec(Buffer, NumSamples*Channels);
  end;
end;

procedure TDCParamEQ.Process24Bit(Buffer: P24BitSample; NumSamples: Integer; Channels: integer);
var
  c, n: integer;
  i: integer;
  so, si: Single;
  m1, m2, m3, m4 : PSingle;
  n1: PSingle;
  pa: Single;
begin
  dec(Buffer);
  for c := 0 to Channels -1 do
  begin
    inc(Buffer);
    pa := FChannelPreAmp[c]^;

    m1 := @FXX[c,0,0];
    m2 := @FXX[c,0,1];
    m3 := @FYY[c,0,0];
    m4 := @FYY[c,0,1];

    n1 := @FB0[c,0];

    for i := 0 to NumSamples -1 do
    begin
      si := Cvt24BitTo32(Buffer^) / 8388608 * pa;

      if (si = 0.0)
        then si := 0.00000001;

      for n := 0 to EQ_NUM_BANDS -1 do
      begin

        so := (FBB[c,n,0]*si + n1^*m1^ + FBB[c,n,1]*m2^ - n1^*m3^ - FAA[c,n,1]*m4^) * FAA[c,n,0];

        m2^ := m1^;
        inc(m2, 2);

        m1^ := si;
        inc(m1, 2);

        m4^ := m3^;
        inc(m4, 2);

        m3^ := so;
        inc(m3, 2);

        inc(n1, 1);

        si := so;
      end;

      dec(m1, EQ_NUM_BANDS_2);
      dec(m2, EQ_NUM_BANDS_2);
      dec(m3, EQ_NUM_BANDS_2);
      dec(m4, EQ_NUM_BANDS_2);
      dec(n1, EQ_NUM_BANDS);

      Buffer^ := Cvt32BitTo24(Clip_24(Round(so * 8388608)));
      inc(Buffer, Channels)
    end;

    dec(Buffer, NumSamples*Channels);
  end;
end;

procedure TDCParamEQ.Process32Bit(Buffer: PInteger; NumSamples: Integer; Channels: integer);
var
  c, n: integer;
  i: integer;
  so, si: Single;
  m1, m2, m3, m4 : PSingle;
  n1: PSingle;
  pa: Single;
begin
  dec(Buffer);
  for c := 0 to Channels -1 do
  begin
    inc(Buffer);
    pa := FChannelPreAmp[c]^;

    m1 := @FXX[c,0,0];
    m2 := @FXX[c,0,1];
    m3 := @FYY[c,0,0];
    m4 := @FYY[c,0,1];

    n1 := @FB0[c,0];

    for i := 0 to NumSamples -1 do
    begin
      si := Buffer^ / 2147483648 * pa;

      if (si = 0.0)
        then si := 0.00000001;

      for n := 0 to EQ_NUM_BANDS -1 do
      begin

        so := (FBB[c,n,0]*si + n1^*m1^ + FBB[c,n,1]*m2^ - n1^*m3^ - FAA[c,n,1]*m4^) * FAA[c,n,0];

        m2^ := m1^;
        inc(m2, 2);

        m1^ := si;
        inc(m1, 2);

        m4^ := m3^;
        inc(m4, 2);

        m3^ := so;
        inc(m3, 2);

        inc(n1, 1);

        si := so;
      end;

      dec(m1, EQ_NUM_BANDS_2);
      dec(m2, EQ_NUM_BANDS_2);
      dec(m3, EQ_NUM_BANDS_2);
      dec(m4, EQ_NUM_BANDS_2);
      dec(n1, EQ_NUM_BANDS);

      Buffer^ := Clip_32(Round(so * 2147483648));
      inc(Buffer, Channels)
    end;

    dec(Buffer, NumSamples*Channels);
  end;
end;

procedure TDCParamEQ.Process32BitFloat(Buffer: PSingle; NumSamples: Integer; Channels: integer);
var
  c, n: integer;
  i: integer;
  so, si: Single;
  m1, m2, m3, m4 : PSingle;
  n1: PSingle;
  pa: Single;
begin
  dec(Buffer);
  for c := 0 to Channels -1 do
  begin
    inc(Buffer);
    pa := FChannelPreAmp[c]^;

    m1 := @FXX[c,0,0];
    m2 := @FXX[c,0,1];
    m3 := @FYY[c,0,0];
    m4 := @FYY[c,0,1];

    n1 := @FB0[c,0];

    for i := 0 to NumSamples -1 do
    begin
      si := Buffer^ * pa;

      if (si = 0.0)
        then si := 0.00000001;

      for n := 0 to EQ_NUM_BANDS -1 do
      begin

        so := (FBB[c,n,0]*si + n1^*m1^ + FBB[c,n,1]*m2^ - n1^*m3^ - FAA[c,n,1]*m4^) * FAA[c,n,0];

        m2^ := m1^;
        inc(m2, 2);

        m1^ := si;
        inc(m1, 2);

        m4^ := m3^;
        inc(m4, 2);

        m3^ := so;
        inc(m3, 2);

        inc(n1, 1);

        si := so;
      end;

      dec(m1, EQ_NUM_BANDS_2);
      dec(m2, EQ_NUM_BANDS_2);
      dec(m3, EQ_NUM_BANDS_2);
      dec(m4, EQ_NUM_BANDS_2);
      dec(n1, EQ_NUM_BANDS);

      Buffer^ := Clip_32F(so);
      inc(Buffer, Channels)
    end;

    dec(Buffer, NumSamples*Channels);
  end;
end;

procedure TDCParamEQ.Process8BitSSE(Buffer: PByte; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
const
  MI: Single = -128.0;
  MA: Single = 127.0;
  BITRATE: Integer = 8 shr 3;
asm
  // edx      Buffer
  // ecx      NumSamples
  // [ebp+20] Channels
  // [ebp+16] Amp
  // [ebp+12] X
  // [ebp+8]  A

  push ebx
  push esi
  push edi

  sub esp, 12

  // Speichert den Buffer Zeiger auf dem Stack [ESP+0]
  mov [esp], edx
  // Speichert NumSamples auf dem Stack [ESP+4]
  mov [esp+4], ecx

  // speichert die Anzahl der Bytes die bei "sampleLoop" zu
  // inkrementieren sind in [ESP+8]
  mov eax, [ebp+20]
  // multiplizieren mit der BitRate/8
  imul eax, BITRATE
  mov [esp+8], eax

  // Speichern des Invertierungs Arrays in XMM7
  // wird benutzt um das Vorzeichen der 2 niederen Werte zu invertieren
  movups xmm7, InvertSSE_LL

@startLoop:
  // dekrementiert den aktuellen Kanal
  dec [ebp+20]

  // speichert FChannelPreAmp Zeiger in ESI
  mov esi, [ebp+16]
  // aktuellen Kanal laden
  mov ebx, [ebp+20]
  // multiplizieren mit SizeOf(Single)
  shl ebx, 2
  // aktuelle Amplifikation laden anhand der Kanal position
  mov esi, [esi+ebx]

  // ldt aktuelle Amplifikation in XMM0
  movss xmm0, DWORD PTR [esi]

  // laden von X in ESI basierend auf Kanal position
  mov esi, [ebp+12]
  mov esi, [esi]
  mov ebx, [ebp+20]
  imul ebx, EQ_NUM_BANDS_ALIGNED_SIZE
  add esi, ebx

  // laden von A in EDX basierend auf Kanal position
  mov edx, [ebp+8]
  mov edx, [edx]
  add edx, ebx

  // laden des Buffer Zeigers in EDI anhand der aktuellen
  // Kanal Position von [EBP+20]
  mov edi, [esp]
  mov ebx, [ebp+20]
  // multiplizieren mit BitRate/8
  imul ebx, BITRATE
  // Zeigt auf Buffer+aktueller Channel
  add edi, ebx

  // Zhler EBX auf letzte Position des Buffers setzen
  mov ebx, [esp+8]
  imul ebx, [esp+4]
  add ebx, edi

@sampleLoop:
  mov cl, [edi]
  add cl, -128
  // Aus Preformance Grnden, lieber mit 0.00000001 anstatt mit 0 multiplizieren
  // hat keine Auswirkung auf das Endergebniss
  cmp cl, 0
  jne @sampleSkip
  movss xmm1, den
  jmp @sampleDo

@sampleSkip:
  movsx ecx, cl
  cvtsi2ss xmm1, ecx

@sampleDo:
  mulss xmm1, xmm0

  // ESI FX
  // EDX FA

  // so := ( FA[c][n][5] * si +
  //         FA[c][n][0] * FX[c][n][0] +
  //         FA[c][n][1] * FX[c][n][1] -
  //         FA[c][n][2] * FX[c][n][2] -
  //         FA[c][n][3] * FX[c][n][3] )
  //       * FA[c][n][4];
  // FX[c][n][1] := FX[c][n][0];
  // FX[c][n][0] := si;
  // FX[c][n][3] := FX[c][n][2];
  // FX[c][n][2] := so;
  // si := so;

  mov ecx, EQ_NUM_BANDS
  shl ecx, 5
  add ecx, -32

@bandLoop:
  // EQ START

  movaps xmm3, DQWORD PTR [edx+ecx]
  movaps xmm5, DQWORD PTR [esi+ecx]
  mulps xmm3, xmm5
  movss xmm2, DWORD PTR [edx+ecx+20]
  mulss xmm2, xmm1
  xorps xmm3, xmm7
  movss [esi+ecx+4], xmm5

  movaps xmm4, xmm3
  shufps xmm4, xmm3, $0E
  addps xmm3, xmm4
  movss xmm4, xmm3
  shufps xmm3, xmm3, 1
  addss xmm3, xmm4

  addss xmm2, xmm3
  movss [esi+ecx], xmm1
  mulss xmm2, DWORD PTR [edx+ecx+16]
  mov eax, [esi+ecx+8]
  mov [esi+ecx+12], eax
  movss [esi+ecx+8], xmm2
  movss xmm1, xmm2

  // EQ ENDE

  add ecx, -32
  cmp ecx, -32
  jb @bandLoop

  // Range Check
  minss xmm1, MA
  maxss xmm1, MI

  // zurckschreiben
  cvtss2si ecx, xmm1
  add cl, 128
  mov [edi], cl

  add edi, [esp+8]
  cmp edi, ebx
  jne @sampleLoop

  cmp [ebp+20], 0
  jne @startLoop

  add esp, 12

  pop edi
  pop esi
  pop ebx
end;

procedure TDCParamEQ.Process16BitSSE(Buffer: PSmallInt; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
const
  MI: Single = -32768.0;
  MA: Single = 32767.0;
  BITRATE: Integer = 16 shr 3;
asm
  // edx      Buffer
  // ecx      NumSamples
  // [ebp+20] Channels
  // [ebp+16] Amp
  // [ebp+12] X
  // [ebp+8]  A

  push ebx
  push esi
  push edi

  sub esp, 12

  // Speichert den Buffer Zeiger auf dem Stack [ESP+0]
  mov [esp], edx
  // Speichert NumSamples auf dem Stack [ESP+4]
  mov [esp+4], ecx

  // speichert die Anzahl der Bytes die bei "sampleLoop" zu
  // inkrementieren sind in [ESP+8]
  mov eax, [ebp+20]
  // multiplizieren mit der BitRate/8
  imul eax, BITRATE
  mov [esp+8], eax

  // Speichern des Invertierungs Arrays in XMM7
  // wird benutzt um das Vorzeichen der 2 niederen Werte zu invertieren
  movups xmm7, InvertSSE_LL

@startLoop:
  // dekrementiert den aktuellen Kanal
  dec [ebp+20]

  // speichert FChannelPreAmp Zeiger in ESI
  mov esi, [ebp+16]
  // aktuellen Kanal laden
  mov ebx, [ebp+20]
  // multiplizieren mit SizeOf(Single)
  shl ebx, 2
  // aktuelle Amplifikation laden anhand der Kanal position
  mov esi, [esi+ebx]

  // ldt aktuelle Amplifikation in XMM0
  movss xmm0, DWORD PTR [esi]

  // laden von X in ESI basierend auf Kanal position
  mov esi, [ebp+12]
  mov esi, [esi]
  mov ebx, [ebp+20]
  imul ebx, EQ_NUM_BANDS_ALIGNED_SIZE
  add esi, ebx

  // laden von A in EDX basierend auf Kanal position
  mov edx, [ebp+8]
  mov edx, [edx]
  add edx, ebx

  // laden des Buffer Zeigers in EDI anhand der aktuellen
  // Kanal Position von [EBP+20]
  mov edi, [esp]
  mov ebx, [ebp+20]
  // multiplizieren mit BitRate/8
  imul ebx, BITRATE
  // Zeigt auf Buffer+aktueller Channel
  add edi, ebx

  // Zhler EBX auf letzte Position des Buffers setzen
  mov ebx, [esp+8]
  imul ebx, [esp+4]
  add ebx, edi

@sampleLoop:
  mov cx, [edi]
  // Aus Preformance Grnden, lieber mit 0.00000001 anstatt mit 0 multiplizieren
  // hat keine Auswirkung auf das Endergebniss
  cmp cx, 0
  jne @sampleSkip
  movss xmm1, den
  jmp @sampleDo

@sampleSkip:
  movsx ecx, cx
  cvtsi2ss xmm1, ecx

@sampleDo:
  mulss xmm1, xmm0

  // ESI FX
  // EDX FA

  // so := ( FA[c][n][5] * si +
  //         FA[c][n][0] * FX[c][n][0] +
  //         FA[c][n][1] * FX[c][n][1] -
  //         FA[c][n][2] * FX[c][n][2] -
  //         FA[c][n][3] * FX[c][n][3] )
  //       * FA[c][n][4];
  // FX[c][n][1] := FX[c][n][0];
  // FX[c][n][0] := si;
  // FX[c][n][3] := FX[c][n][2];
  // FX[c][n][2] := so;
  // si := so;

  mov ecx, EQ_NUM_BANDS
  shl ecx, 5
  add ecx, -32

@bandLoop:
  // EQ START

  movaps xmm3, DQWORD PTR [edx+ecx]
  movaps xmm5, DQWORD PTR [esi+ecx]
  mulps xmm3, xmm5
  movss xmm2, DWORD PTR [edx+ecx+20]
  mulss xmm2, xmm1
  xorps xmm3, xmm7
  movss [esi+ecx+4], xmm5

  movaps xmm4, xmm3
  shufps xmm4, xmm3, $0E
  addps xmm3, xmm4
  movss xmm4, xmm3
  shufps xmm3, xmm3, 1
  addss xmm3, xmm4

  addss xmm2, xmm3
  movss [esi+ecx], xmm1
  mulss xmm2, DWORD PTR [edx+ecx+16]
  mov eax, [esi+ecx+8]
  mov [esi+ecx+12], eax
  movss [esi+ecx+8], xmm2
  movss xmm1, xmm2

  // EQ ENDE

  add ecx, -32
  cmp ecx, -32
  jb @bandLoop

  // Range Check
  minss xmm1, MA
  maxss xmm1, MI

  // zurckschreiben
  cvtss2si ecx, xmm1
  mov [edi], cx

  add edi, [esp+8]
  cmp edi, ebx
  jne @sampleLoop

  cmp [ebp+20], 0
  jne @startLoop

  add esp, 12

  pop edi
  pop esi
  pop ebx
end;

procedure TDCParamEQ.Process24BitSSE(Buffer: P24BitSample; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
const
  MI: Single = -8388608.0;
  MA: Single = 8388607.0;
  BITRATE: Integer = 24 shr 3;
asm
  // edx      Buffer
  // ecx      NumSamples
  // [ebp+20] Channels
  // [ebp+16] Amp
  // [ebp+12] X
  // [ebp+8]  A

  push ebx
  push esi
  push edi

  sub esp, 12

  // Speichert den Buffer Zeiger auf dem Stack [ESP+0]
  mov [esp], edx
  // Speichert NumSamples auf dem Stack [ESP+4]
  mov [esp+4], ecx

  // speichert die Anzahl der Bytes die bei "sampleLoop" zu
  // inkrementieren sind in [ESP+8]
  mov eax, [ebp+20]
  // multiplizieren mit der BitRate/8
  imul eax, BITRATE
  mov [esp+8], eax

  // Speichern des Invertierungs Arrays in XMM7
  // wird benutzt um das Vorzeichen der 2 niederen Werte zu invertieren
  movups xmm7, InvertSSE_LL

@startLoop:
  // dekrementiert den aktuellen Kanal
  dec [ebp+20]

  // speichert FChannelPreAmp Zeiger in ESI
  mov esi, [ebp+16]
  // aktuellen Kanal laden
  mov ebx, [ebp+20]
  // multiplizieren mit SizeOf(Single)
  shl ebx, 2
  // aktuelle Amplifikation laden anhand der Kanal position
  mov esi, [esi+ebx]

  // ldt aktuelle Amplifikation in XMM0
  movss xmm0, DWORD PTR [esi]

  // laden von X in ESI basierend auf Kanal position
  mov esi, [ebp+12]
  mov esi, [esi]
  mov ebx, [ebp+20]
  imul ebx, EQ_NUM_BANDS_ALIGNED_SIZE
  add esi, ebx

  // laden von A in EDX basierend auf Kanal position
  mov edx, [ebp+8]
  mov edx, [edx]
  add edx, ebx

  // laden des Buffer Zeigers in EDI anhand der aktuellen
  // Kanal Position von [EBP+20]
  mov edi, [esp]
  mov ebx, [ebp+20]
  // multiplizieren mit BitRate/8
  imul ebx, BITRATE
  // Zeigt auf Buffer+aktueller Channel
  add edi, ebx

  // Zhler EBX auf letzte Position des Buffers setzen
  mov ebx, [esp+8]
  imul ebx, [esp+4]
  add ebx, edi

@sampleLoop:
  mov ecx, [edi]
  shl ecx, 8
  sar ecx, 8
  // Aus Preformance Grnden, lieber mit 0.00000001 anstatt mit 0 multiplizieren
  // hat keine Auswirkung auf das Endergebniss
  cmp ecx, 0
  jne @sampleSkip
  movss xmm1, den
  jmp @sampleDo

@sampleSkip:
  cvtsi2ss xmm1, ecx

@sampleDo:
  mulss xmm1, xmm0

  // ESI FX
  // EDX FA

  // so := ( FA[c][n][5] * si +
  //         FA[c][n][0] * FX[c][n][0] +
  //         FA[c][n][1] * FX[c][n][1] -
  //         FA[c][n][2] * FX[c][n][2] -
  //         FA[c][n][3] * FX[c][n][3] )
  //       * FA[c][n][4];
  // FX[c][n][1] := FX[c][n][0];
  // FX[c][n][0] := si;
  // FX[c][n][3] := FX[c][n][2];
  // FX[c][n][2] := so;
  // si := so;

  mov ecx, EQ_NUM_BANDS
  shl ecx, 5
  add ecx, -32

@bandLoop:
  // EQ START

  movaps xmm3, DQWORD PTR [edx+ecx]
  movaps xmm5, DQWORD PTR [esi+ecx]
  mulps xmm3, xmm5
  movss xmm2, DWORD PTR [edx+ecx+20]
  mulss xmm2, xmm1
  xorps xmm3, xmm7
  movss [esi+ecx+4], xmm5

  movaps xmm4, xmm3
  shufps xmm4, xmm3, $0E
  addps xmm3, xmm4
  movss xmm4, xmm3
  shufps xmm3, xmm3, 1
  addss xmm3, xmm4

  addss xmm2, xmm3
  movss [esi+ecx], xmm1
  mulss xmm2, DWORD PTR [edx+ecx+16]
  mov eax, [esi+ecx+8]
  mov [esi+ecx+12], eax
  movss [esi+ecx+8], xmm2
  movss xmm1, xmm2

  // EQ ENDE

  add ecx, -32
  cmp ecx, -32
  jb @bandLoop

  // Range Check
  minss xmm1, MA
  maxss xmm1, MI

  // zurckschreiben
  cvtss2si ecx, xmm1

  mov BYTE [edi], cl
  sar ecx, 8
  mov WORD [edi+1], cx

  add edi, [esp+8]
  cmp edi, ebx
  jne @sampleLoop

  cmp [ebp+20], 0
  jne @startLoop

  add esp, 12

  pop edi
  pop esi
  pop ebx
end;

procedure TDCParamEQ.Process32BitIntegerSSE(Buffer: PInteger; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
const
  MI: Single = -2147483648.0;
  MA: Single = 2147483647.0;
  BITRATE: Integer = 32 shr 3;
asm
  // edx      Buffer
  // ecx      NumSamples
  // [ebp+20] Channels
  // [ebp+16] Amp
  // [ebp+12] X
  // [ebp+8]  A

  push ebx
  push esi
  push edi

  sub esp, 12

  // Speichert den Buffer Zeiger auf dem Stack [ESP+0]
  mov [esp], edx
  // Speichert NumSamples auf dem Stack [ESP+4]
  mov [esp+4], ecx

  // speichert die Anzahl der Bytes die bei "sampleLoop" zu
  // inkrementieren sind in [ESP+8]
  mov eax, [ebp+20]
  // multiplizieren mit der BitRate/8
  imul eax, BITRATE
  mov [esp+8], eax

  // Speichern des Invertierungs Arrays in XMM7
  // wird benutzt um das Vorzeichen der 2 niederen Werte zu invertieren
  movups xmm7, InvertSSE_LL

@startLoop:
  // dekrementiert den aktuellen Kanal
  dec [ebp+20]

  // speichert FChannelPreAmp Zeiger in ESI
  mov esi, [ebp+16]
  // aktuellen Kanal laden
  mov ebx, [ebp+20]
  // multiplizieren mit SizeOf(Single)
  shl ebx, 2
  // aktuelle Amplifikation laden anhand der Kanal position
  mov esi, [esi+ebx]

  // ldt aktuelle Amplifikation in XMM0
  movss xmm0, DWORD PTR [esi]

  // laden von X in ESI basierend auf Kanal position
  mov esi, [ebp+12]
  mov esi, [esi]
  mov ebx, [ebp+20]
  imul ebx, EQ_NUM_BANDS_ALIGNED_SIZE
  add esi, ebx

  // laden von A in EDX basierend auf Kanal position
  mov edx, [ebp+8]
  mov edx, [edx]
  add edx, ebx

  // laden des Buffer Zeigers in EDI anhand der aktuellen
  // Kanal Position von [EBP+20]
  mov edi, [esp]
  mov ebx, [ebp+20]
  // multiplizieren mit BitRate/8
  imul ebx, BITRATE
  // Zeigt auf Buffer+aktueller Channel
  add edi, ebx

  // Zhler EBX auf letzte Position des Buffers setzen
  mov ebx, [esp+8]
  imul ebx, [esp+4]
  add ebx, edi

@sampleLoop:
  // Aus Preformance Grnden, lieber mit 0.00000001 anstatt mit 0 multiplizieren
  // hat keine Auswirkung auf das Endergebniss
  cmp [edi], 0
  jne @sampleSkip
  movss xmm1, den
  jmp @sampleDo

@sampleSkip:
  cvtsi2ss xmm1, [edi]

@sampleDo:
  mulss xmm1, xmm0

  // ESI FX
  // EDX FA

  // so := ( FA[c][n][5] * si +
  //         FA[c][n][0] * FX[c][n][0] +
  //         FA[c][n][1] * FX[c][n][1] -
  //         FA[c][n][2] * FX[c][n][2] -
  //         FA[c][n][3] * FX[c][n][3] )
  //       * FA[c][n][4];
  // FX[c][n][1] := FX[c][n][0];
  // FX[c][n][0] := si;
  // FX[c][n][3] := FX[c][n][2];
  // FX[c][n][2] := so;
  // si := so;

  mov ecx, EQ_NUM_BANDS
  shl ecx, 5
  add ecx, -32

@bandLoop:
  // EQ START

  movaps xmm3, DQWORD PTR [edx+ecx]
  movaps xmm5, DQWORD PTR [esi+ecx]
  mulps xmm3, xmm5
  movss xmm2, DWORD PTR [edx+ecx+20]
  mulss xmm2, xmm1
  xorps xmm3, xmm7
  movss [esi+ecx+4], xmm5

  movaps xmm4, xmm3
  shufps xmm4, xmm3, $0E
  addps xmm3, xmm4
  movss xmm4, xmm3
  shufps xmm3, xmm3, 1
  addss xmm3, xmm4

  addss xmm2, xmm3
  movss [esi+ecx], xmm1
  mulss xmm2, DWORD PTR [edx+ecx+16]
  mov eax, [esi+ecx+8]
  mov [esi+ecx+12], eax
  movss [esi+ecx+8], xmm2
  movss xmm1, xmm2

  // EQ ENDE

  add ecx, -32
  cmp ecx, -32
  jb @bandLoop

  // Range Check
  minss xmm1, MA
  maxss xmm1, MI

  // zurckschreiben
  cvtss2si ecx, xmm1
  mov [edi], ecx

  add edi, [esp+8]
  cmp edi, ebx
  jne @sampleLoop

  cmp [ebp+20], 0
  jne @startLoop

  add esp, 12

  pop edi
  pop esi
  pop ebx
end;

procedure TDCParamEQ.Process32BitFloatSSE(Buffer: PSingle; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
const
  MI: Single = -1.0;
  MA: Single = 1.0;
  BITRATE: Integer = 32 shr 3;
asm
  // edx      Buffer
  // ecx      NumSamples
  // [ebp+20] Channels
  // [ebp+16] Amp
  // [ebp+12] X
  // [ebp+8]  A

  push ebx
  push esi
  push edi

  sub esp, 12

  // Speichert den Buffer Zeiger auf dem Stack [ESP+0]
  mov [esp], edx
  // Speichert NumSamples auf dem Stack [ESP+4]
  mov [esp+4], ecx

  // speichert die Anzahl der Bytes die bei "sampleLoop" zu
  // inkrementieren sind in [ESP+8]
  mov eax, [ebp+20]
  // multiplizieren mit der BitRate/8
  imul eax, BITRATE
  mov [esp+8], eax

  // Speichern des Invertierungs Arrays in XMM7
  // wird benutzt um das Vorzeichen der 2 niederen Werte zu invertieren
  movups xmm7, InvertSSE_LL

@startLoop:
  // dekrementiert den aktuellen Kanal
  dec [ebp+20]

  // speichert FChannelPreAmp Zeiger in ESI
  mov esi, [ebp+16]
  // aktuellen Kanal laden
  mov ebx, [ebp+20]
  // multiplizieren mit SizeOf(Single)
  shl ebx, 2
  // aktuelle Amplifikation laden anhand der Kanal position
  mov esi, [esi+ebx]

  // ldt aktuelle Amplifikation in XMM0
  movss xmm0, DWORD PTR [esi]

  // laden von X in ESI basierend auf Kanal position
  mov esi, [ebp+12]
  mov esi, [esi]
  mov ebx, [ebp+20]
  imul ebx, EQ_NUM_BANDS_ALIGNED_SIZE
  add esi, ebx

  // laden von A in EDX basierend auf Kanal position
  mov edx, [ebp+8]
  mov edx, [edx]
  add edx, ebx

  // laden des Buffer Zeigers in EDI anhand der aktuellen
  // Kanal Position von [EBP+20]
  mov edi, [esp]
  mov ebx, [ebp+20]
  // multiplizieren mit BitRate/8
  imul ebx, BITRATE
  // Zeigt auf Buffer+aktueller Channel
  add edi, ebx

  // Zhler EBX auf letzte Position des Buffers setzen
  mov ebx, [esp+8]
  imul ebx, [esp+4]
  add ebx, edi

@sampleLoop:
  // Aus Preformance Grnden, lieber mit 0.00000001 anstatt mit 0 multiplizieren
  // hat keine Auswirkung auf das Endergebniss
  cmp [edi], 0
  jne @sampleSkip
  movss xmm1, den
  jmp @sampleDo

@sampleSkip:
  movss xmm1, DWORD PTR [edi]

@sampleDo:
  mulss xmm1, xmm0

  // ESI FX
  // EDX FA

  // so := ( FA[c][n][5] * si +
  //         FA[c][n][0] * FX[c][n][0] +
  //         FA[c][n][1] * FX[c][n][1] -
  //         FA[c][n][2] * FX[c][n][2] -
  //         FA[c][n][3] * FX[c][n][3] )
  //       * FA[c][n][4];
  // FX[c][n][1] := FX[c][n][0];
  // FX[c][n][0] := si;
  // FX[c][n][3] := FX[c][n][2];
  // FX[c][n][2] := so;
  // si := so;

  mov ecx, EQ_NUM_BANDS
  shl ecx, 5
  add ecx, -32

@bandLoop:
  // EQ START

  movaps xmm3, DQWORD PTR [edx+ecx]
  movaps xmm5, DQWORD PTR [esi+ecx]
  mulps xmm3, xmm5
  movss xmm2, DWORD PTR [edx+ecx+20]
  mulss xmm2, xmm1
  xorps xmm3, xmm7
  movss [esi+ecx+4], xmm5

  movaps xmm4, xmm3
  shufps xmm4, xmm3, $0E
  addps xmm3, xmm4
  movss xmm4, xmm3
  shufps xmm3, xmm3, 1
  addss xmm3, xmm4

  addss xmm2, xmm3
  movss [esi+ecx], xmm1
  mulss xmm2, DWORD PTR [edx+ecx+16]
  mov eax, [esi+ecx+8]
  mov [esi+ecx+12], eax
  movss [esi+ecx+8], xmm2
  movss xmm1, xmm2

  // EQ ENDE

  add ecx, -32
  cmp ecx, -32
  jb @bandLoop

  // Range Check
  minss xmm1, MA
  maxss xmm1, MI

  // zurckschreiben
  movss [edi] ,xmm1

  add edi, [esp+8]
  cmp edi, ebx
  jne @sampleLoop

  cmp [ebp+20], 0
  jne @startLoop

  add esp, 12

  pop edi
  pop esi
  pop ebx
end;

procedure TDCParamEQ.Process8BitSSE3(Buffer: PByte; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
const
  MI: Single = -128.0;
  MA: Single = 127.0;
  BITRATE: Integer = 8 shr 3;
asm
  // edx      Buffer
  // ecx      NumSamples
  // [ebp+20] Channels
  // [ebp+16] Amp
  // [ebp+12] X
  // [ebp+8]  A

  push ebx
  push esi
  push edi

  sub esp, 12

  // Speichert den Buffer Zeiger auf dem Stack [ESP+0]
  mov [esp], edx
  // Speichert NumSamples auf dem Stack [ESP+4]
  mov [esp+4], ecx

  // speichert die Anzahl der Bytes die bei "sampleLoop" zu
  // inkrementieren sind in [ESP+8]
  mov eax, [ebp+20]
  // multiplizieren mit der BitRate/8
  imul eax, BITRATE
  mov [esp+8], eax

  // Speichern des Invertierungs Arrays in XMM7
  // wird benutzt um das Vorzeichen der 2 niederen Werte zu invertieren
  movups xmm7, InvertSSE_LL

@startLoop:
  // dekrementiert den aktuellen Kanal
  dec [ebp+20]

  // speichert FChannelPreAmp Zeiger in ESI
  mov esi, [ebp+16]
  // aktuellen Kanal laden
  mov ebx, [ebp+20]
  // multiplizieren mit SizeOf(Single)
  shl ebx, 2
  // aktuelle Amplifikation laden anhand der Kanal position
  mov esi, [esi+ebx]

  // ldt aktuelle Amplifikation in XMM0
  movss xmm0, DWORD PTR [esi]

  // laden von X in ESI basierend auf Kanal position
  mov esi, [ebp+12]
  mov esi, [esi]
  mov ebx, [ebp+20]
  imul ebx, EQ_NUM_BANDS_ALIGNED_SIZE
  add esi, ebx

  // laden von A in EDX basierend auf Kanal position
  mov edx, [ebp+8]
  mov edx, [edx]
  add edx, ebx

  // laden des Buffer Zeigers in EDI anhand der aktuellen
  // Kanal Position von [EBP+20]
  mov edi, [esp]
  mov ebx, [ebp+20]
  // multiplizieren mit BitRate/8
  imul ebx, BITRATE
  // Zeigt auf Buffer+aktueller Channel
  add edi, ebx

  // Zhler EBX auf letzte Position des Buffers setzen
  mov ebx, [esp+8]
  imul ebx, [esp+4]
  add ebx, edi

@sampleLoop:
  mov cl, [edi]
  add cl, -128
  // Aus Preformance Grnden, lieber mit 0.00000001 anstatt mit 0 multiplizieren
  // hat keine Auswirkung auf das Endergebniss
  cmp cl, 0
  jne @sampleSkip
  movss xmm1, den
  jmp @sampleDo

@sampleSkip:
  movsx ecx, cl
  cvtsi2ss xmm1, ecx

@sampleDo:
  mulss xmm1, xmm0

  // ESI FX
  // EDX FA

  // so := ( FA[c][n][5] * si +
  //         FA[c][n][0] * FX[c][n][0] +
  //         FA[c][n][1] * FX[c][n][1] -
  //         FA[c][n][2] * FX[c][n][2] -
  //         FA[c][n][3] * FX[c][n][3] )
  //       * FA[c][n][4];
  // FX[c][n][1] := FX[c][n][0];
  // FX[c][n][0] := si;
  // FX[c][n][3] := FX[c][n][2];
  // FX[c][n][2] := so;
  // si := so;

  mov ecx, EQ_NUM_BANDS
  shl ecx, 5
  add ecx, -32

@bandLoop:
  // EQ START

  movaps xmm3, DQWORD PTR [edx+ecx]
  movaps xmm5, DQWORD PTR [esi+ecx]
  mulps xmm3, xmm5
  movss xmm2, DWORD PTR [edx+ecx+20]
  mulss xmm2, xmm1
  xorps xmm3, xmm7
  movss [esi+ecx+4], xmm5

  db $f2,$0f,$7c,$db   // HADDPS xmm3, xmm3
  db $f2,$0f,$7c,$db   // HADDPS xmm3, xmm3

  addss xmm2, xmm3
  movss [esi+ecx], xmm1
  mulss xmm2, DWORD PTR [edx+ecx+16]
  mov eax, [esi+ecx+8]
  mov [esi+ecx+12], eax
  movss [esi+ecx+8], xmm2
  movss xmm1, xmm2

  // EQ ENDE

  add ecx, -32
  cmp ecx, -32
  jb @bandLoop

  // Range Check
  minss xmm1, MA
  maxss xmm1, MI

  // zurckschreiben
  cvtss2si ecx, xmm1
  add cl, 128
  mov [edi], cl

  add edi, [esp+8]
  cmp edi, ebx
  jne @sampleLoop

  cmp [ebp+20], 0
  jne @startLoop

  add esp, 12

  pop edi
  pop esi
  pop ebx
end;

procedure TDCParamEQ.Process16BitSSE3(Buffer: PSmallInt; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
const
  MI: Single = -32768.0;
  MA: Single = 32767.0;
  BITRATE: Integer = 16 shr 3;
asm
  // edx      Buffer
  // ecx      NumSamples
  // [ebp+20] Channels
  // [ebp+16] Amp
  // [ebp+12] X
  // [ebp+8]  A

  push ebx
  push esi
  push edi

  sub esp, 12

  // Speichert den Buffer Zeiger auf dem Stack [ESP+0]
  mov [esp], edx
  // Speichert NumSamples auf dem Stack [ESP+4]
  mov [esp+4], ecx

  // speichert die Anzahl der Bytes die bei "sampleLoop" zu
  // inkrementieren sind in [ESP+8]
  mov eax, [ebp+20]
  // multiplizieren mit der BitRate/8
  imul eax, BITRATE
  mov [esp+8], eax

  // Speichern des Invertierungs Arrays in XMM7
  // wird benutzt um das Vorzeichen der 2 niederen Werte zu invertieren
  movups xmm7, InvertSSE_LL

@startLoop:
  // dekrementiert den aktuellen Kanal
  dec [ebp+20]

  // speichert FChannelPreAmp Zeiger in ESI
  mov esi, [ebp+16]
  // aktuellen Kanal laden
  mov ebx, [ebp+20]
  // multiplizieren mit SizeOf(Single)
  shl ebx, 2
  // aktuelle Amplifikation laden anhand der Kanal position
  mov esi, [esi+ebx]

  // ldt aktuelle Amplifikation in XMM0
  movss xmm0, DWORD PTR [esi]

  // laden von X in ESI basierend auf Kanal position
  mov esi, [ebp+12]
  mov esi, [esi]
  mov ebx, [ebp+20]
  imul ebx, EQ_NUM_BANDS_ALIGNED_SIZE
  add esi, ebx

  // laden von A in EDX basierend auf Kanal position
  mov edx, [ebp+8]
  mov edx, [edx]
  add edx, ebx

  // laden des Buffer Zeigers in EDI anhand der aktuellen
  // Kanal Position von [EBP+20]
  mov edi, [esp]
  mov ebx, [ebp+20]
  // multiplizieren mit BitRate/8
  imul ebx, BITRATE
  // Zeigt auf Buffer+aktueller Channel
  add edi, ebx

  // Zhler EBX auf letzte Position des Buffers setzen
  mov ebx, [esp+8]
  imul ebx, [esp+4]
  add ebx, edi

@sampleLoop:
  mov cx, [edi]
  // Aus Preformance Grnden, lieber mit 0.00000001 anstatt mit 0 multiplizieren
  // hat keine Auswirkung auf das Endergebniss
  cmp cx, 0
  jne @sampleSkip
  movss xmm1, den
  jmp @sampleDo

@sampleSkip:
  movsx ecx, cx
  cvtsi2ss xmm1, ecx

@sampleDo:
  mulss xmm1, xmm0

  // ESI FX
  // EDX FA

  // so := ( FA[c][n][5] * si +
  //         FA[c][n][0] * FX[c][n][0] +
  //         FA[c][n][1] * FX[c][n][1] -
  //         FA[c][n][2] * FX[c][n][2] -
  //         FA[c][n][3] * FX[c][n][3] )
  //       * FA[c][n][4];
  // FX[c][n][1] := FX[c][n][0];
  // FX[c][n][0] := si;
  // FX[c][n][3] := FX[c][n][2];
  // FX[c][n][2] := so;
  // si := so;

  mov ecx, EQ_NUM_BANDS
  shl ecx, 5
  add ecx, -32

@bandLoop:
  // EQ START

  movaps xmm3, DQWORD PTR [edx+ecx]
  movaps xmm5, DQWORD PTR [esi+ecx]
  mulps xmm3, xmm5
  movss xmm2, DWORD PTR [edx+ecx+20]
  mulss xmm2, xmm1
  xorps xmm3, xmm7
  movss [esi+ecx+4], xmm5

  db $f2,$0f,$7c,$db   // HADDPS xmm3, xmm3
  db $f2,$0f,$7c,$db   // HADDPS xmm3, xmm3

  addss xmm2, xmm3
  movss [esi+ecx], xmm1
  mulss xmm2, DWORD PTR [edx+ecx+16]
  mov eax, [esi+ecx+8]
  mov [esi+ecx+12], eax
  movss [esi+ecx+8], xmm2
  movss xmm1, xmm2

  // EQ ENDE

  add ecx, -32
  cmp ecx, -32
  jb @bandLoop

  // Range Check
  minss xmm1, MA
  maxss xmm1, MI

  // zurckschreiben
  cvtss2si ecx, xmm1
  mov [edi], cx

  add edi, [esp+8]
  cmp edi, ebx
  jne @sampleLoop

  cmp [ebp+20], 0
  jne @startLoop

  add esp, 12

  pop edi
  pop esi
  pop ebx
end;

procedure TDCParamEQ.Process24BitSSE3(Buffer: P24BitSample; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
const
  MI: Single = -8388608.0;
  MA: Single = 8388607.0;
  BITRATE: Integer = 24 shr 3;
asm
  // edx      Buffer
  // ecx      NumSamples
  // [ebp+20] Channels
  // [ebp+16] Amp
  // [ebp+12] X
  // [ebp+8]  A

  push ebx
  push esi
  push edi

  sub esp, 12

  // Speichert den Buffer Zeiger auf dem Stack [ESP+0]
  mov [esp], edx
  // Speichert NumSamples auf dem Stack [ESP+4]
  mov [esp+4], ecx

  // speichert die Anzahl der Bytes die bei "sampleLoop" zu
  // inkrementieren sind in [ESP+8]
  mov eax, [ebp+20]
  // multiplizieren mit der BitRate/8
  imul eax, BITRATE
  mov [esp+8], eax

  // Speichern des Invertierungs Arrays in XMM7
  // wird benutzt um das Vorzeichen der 2 niederen Werte zu invertieren
  movups xmm7, InvertSSE_LL

@startLoop:
  // dekrementiert den aktuellen Kanal
  dec [ebp+20]

  // speichert FChannelPreAmp Zeiger in ESI
  mov esi, [ebp+16]
  // aktuellen Kanal laden
  mov ebx, [ebp+20]
  // multiplizieren mit SizeOf(Single)
  shl ebx, 2
  // aktuelle Amplifikation laden anhand der Kanal position
  mov esi, [esi+ebx]

  // ldt aktuelle Amplifikation in XMM0
  movss xmm0, DWORD PTR [esi]

  // laden von X in ESI basierend auf Kanal position
  mov esi, [ebp+12]
  mov esi, [esi]
  mov ebx, [ebp+20]
  imul ebx, EQ_NUM_BANDS_ALIGNED_SIZE
  add esi, ebx

  // laden von A in EDX basierend auf Kanal position
  mov edx, [ebp+8]
  mov edx, [edx]
  add edx, ebx

  // laden des Buffer Zeigers in EDI anhand der aktuellen
  // Kanal Position von [EBP+20]
  mov edi, [esp]
  mov ebx, [ebp+20]
  // multiplizieren mit BitRate/8
  imul ebx, BITRATE
  // Zeigt auf Buffer+aktueller Channel
  add edi, ebx

  // Zhler EBX auf letzte Position des Buffers setzen
  mov ebx, [esp+8]
  imul ebx, [esp+4]
  add ebx, edi

@sampleLoop:
  mov ecx, [edi]
  shl ecx, 8
  sar ecx, 8
  // Aus Preformance Grnden, lieber mit 0.00000001 anstatt mit 0 multiplizieren
  // hat keine Auswirkung auf das Endergebniss
  cmp ecx, 0
  jne @sampleSkip
  movss xmm1, den
  jmp @sampleDo

@sampleSkip:
  cvtsi2ss xmm1, ecx

@sampleDo:
  mulss xmm1, xmm0

  // ESI FX
  // EDX FA

  // so := ( FA[c][n][5] * si +
  //         FA[c][n][0] * FX[c][n][0] +
  //         FA[c][n][1] * FX[c][n][1] -
  //         FA[c][n][2] * FX[c][n][2] -
  //         FA[c][n][3] * FX[c][n][3] )
  //       * FA[c][n][4];
  // FX[c][n][1] := FX[c][n][0];
  // FX[c][n][0] := si;
  // FX[c][n][3] := FX[c][n][2];
  // FX[c][n][2] := so;
  // si := so;

  mov ecx, EQ_NUM_BANDS
  shl ecx, 5
  add ecx, -32

@bandLoop:
  // EQ START

  movaps xmm3, DQWORD PTR [edx+ecx]
  movaps xmm5, DQWORD PTR [esi+ecx]
  mulps xmm3, xmm5
  movss xmm2, DWORD PTR [edx+ecx+20]
  mulss xmm2, xmm1
  xorps xmm3, xmm7
  movss [esi+ecx+4], xmm5

  db $f2,$0f,$7c,$db   // HADDPS xmm3, xmm3
  db $f2,$0f,$7c,$db   // HADDPS xmm3, xmm3

  addss xmm2, xmm3
  movss [esi+ecx], xmm1
  mulss xmm2, DWORD PTR [edx+ecx+16]
  mov eax, [esi+ecx+8]
  mov [esi+ecx+12], eax
  movss [esi+ecx+8], xmm2
  movss xmm1, xmm2

  // EQ ENDE

  add ecx, -32
  cmp ecx, -32
  jb @bandLoop

  // Range Check
  minss xmm1, MA
  maxss xmm1, MI

  // zurckschreiben
  cvtss2si ecx, xmm1

  mov BYTE [edi], cl
  sar ecx, 8
  mov WORD [edi+1], cx

  add edi, [esp+8]
  cmp edi, ebx
  jne @sampleLoop

  cmp [ebp+20], 0
  jne @startLoop

  add esp, 12

  pop edi
  pop esi
  pop ebx
end;

procedure TDCParamEQ.Process32BitIntegerSSE3(Buffer: PInteger; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
const
  MI: Single = -2147483648.0;
  MA: Single = 2147483647.0;
  BITRATE: Integer = 32 shr 3;
asm
  // edx      Buffer
  // ecx      NumSamples
  // [ebp+20] Channels
  // [ebp+16] Amp
  // [ebp+12] X
  // [ebp+8]  A

  push ebx
  push esi
  push edi

  sub esp, 12

  // Speichert den Buffer Zeiger auf dem Stack [ESP+0]
  mov [esp], edx
  // Speichert NumSamples auf dem Stack [ESP+4]
  mov [esp+4], ecx

  // speichert die Anzahl der Bytes die bei "sampleLoop" zu
  // inkrementieren sind in [ESP+8]
  mov eax, [ebp+20]
  // multiplizieren mit der BitRate/8
  imul eax, BITRATE
  mov [esp+8], eax

  // Speichern des Invertierungs Arrays in XMM7
  // wird benutzt um das Vorzeichen der 2 niederen Werte zu invertieren
  movups xmm7, InvertSSE_LL

@startLoop:
  // dekrementiert den aktuellen Kanal
  dec [ebp+20]

  // speichert FChannelPreAmp Zeiger in ESI
  mov esi, [ebp+16]
  // aktuellen Kanal laden
  mov ebx, [ebp+20]
  // multiplizieren mit SizeOf(Single)
  shl ebx, 2
  // aktuelle Amplifikation laden anhand der Kanal position
  mov esi, [esi+ebx]

  // ldt aktuelle Amplifikation in XMM0
  movss xmm0, DWORD PTR [esi]

  // laden von X in ESI basierend auf Kanal position
  mov esi, [ebp+12]
  mov esi, [esi]
  mov ebx, [ebp+20]
  imul ebx, EQ_NUM_BANDS_ALIGNED_SIZE
  add esi, ebx

  // laden von A in EDX basierend auf Kanal position
  mov edx, [ebp+8]
  mov edx, [edx]
  add edx, ebx

  // laden des Buffer Zeigers in EDI anhand der aktuellen
  // Kanal Position von [EBP+20]
  mov edi, [esp]
  mov ebx, [ebp+20]
  // multiplizieren mit BitRate/8
  imul ebx, BITRATE
  // Zeigt auf Buffer+aktueller Channel
  add edi, ebx

  // Zhler EBX auf letzte Position des Buffers setzen
  mov ebx, [esp+8]
  imul ebx, [esp+4]
  add ebx, edi

@sampleLoop:
  // Aus Preformance Grnden, lieber mit 0.00000001 anstatt mit 0 multiplizieren
  // hat keine Auswirkung auf das Endergebniss
  cmp [edi], 0
  jne @sampleSkip
  movss xmm1, den
  jmp @sampleDo

@sampleSkip:
  cvtsi2ss xmm1, [edi]

@sampleDo:
  mulss xmm1, xmm0

  // ESI FX
  // EDX FA

  // so := ( FA[c][n][5] * si +
  //         FA[c][n][0] * FX[c][n][0] +
  //         FA[c][n][1] * FX[c][n][1] -
  //         FA[c][n][2] * FX[c][n][2] -
  //         FA[c][n][3] * FX[c][n][3] )
  //       * FA[c][n][4];
  // FX[c][n][1] := FX[c][n][0];
  // FX[c][n][0] := si;
  // FX[c][n][3] := FX[c][n][2];
  // FX[c][n][2] := so;
  // si := so;

  mov ecx, EQ_NUM_BANDS
  shl ecx, 5
  add ecx, -32

@bandLoop:
  // EQ START

  movaps xmm3, DQWORD PTR [edx+ecx]
  movaps xmm5, DQWORD PTR [esi+ecx]
  mulps xmm3, xmm5
  movss xmm2, DWORD PTR [edx+ecx+20]
  mulss xmm2, xmm1
  xorps xmm3, xmm7
  movss [esi+ecx+4], xmm5

  db $f2,$0f,$7c,$db   // HADDPS xmm3, xmm3
  db $f2,$0f,$7c,$db   // HADDPS xmm3, xmm3

  addss xmm2, xmm3
  movss [esi+ecx], xmm1
  mulss xmm2, DWORD PTR [edx+ecx+16]
  mov eax, [esi+ecx+8]
  mov [esi+ecx+12], eax
  movss [esi+ecx+8], xmm2
  movss xmm1, xmm2

  // EQ ENDE

  add ecx, -32
  cmp ecx, -32
  jb @bandLoop

  // Range Check
  minss xmm1, MA
  maxss xmm1, MI

  // zurckschreiben
  cvtss2si ecx, xmm1
  mov [edi], ecx

  add edi, [esp+8]
  cmp edi, ebx
  jne @sampleLoop

  cmp [ebp+20], 0
  jne @startLoop

  add esp, 12

  pop edi
  pop esi
  pop ebx
end;

procedure TDCParamEQ.Process32BitFloatSSE3(Buffer: PSingle; NumSamples: Integer; Channels: integer; Amp: PFloatArray; X: PFloatArray; A: PFloatArray);
const
  MI: Single = -1.0;
  MA: Single = 1.0;
  BITRATE: Integer = 32 shr 3;
asm
  // edx      Buffer
  // ecx      NumSamples
  // [ebp+20] Channels
  // [ebp+16] Amp
  // [ebp+12] X
  // [ebp+8]  A

  push ebx
  push esi
  push edi

  sub esp, 12

  // Speichert den Buffer Zeiger auf dem Stack [ESP+0]
  mov [esp], edx
  // Speichert NumSamples auf dem Stack [ESP+4]
  mov [esp+4], ecx

  // speichert die Anzahl der Bytes die bei "sampleLoop" zu
  // inkrementieren sind in [ESP+8]
  mov eax, [ebp+20]
  // multiplizieren mit der BitRate/8
  imul eax, BITRATE
  mov [esp+8], eax

  // Speichern des Invertierungs Arrays in XMM7
  // wird benutzt um das Vorzeichen der 2 niederen Werte zu invertieren
  movups xmm7, InvertSSE_LL

@startLoop:
  // dekrementiert den aktuellen Kanal
  dec [ebp+20]

  // speichert FChannelPreAmp Zeiger in ESI
  mov esi, [ebp+16]
  // aktuellen Kanal laden
  mov ebx, [ebp+20]
  // multiplizieren mit SizeOf(Single)
  shl ebx, 2
  // aktuelle Amplifikation laden anhand der Kanal position
  mov esi, [esi+ebx]

  // ldt aktuelle Amplifikation in XMM0
  movss xmm0, DWORD PTR [esi]

  // laden von X in ESI basierend auf Kanal position
  mov esi, [ebp+12]
  mov esi, [esi]
  mov ebx, [ebp+20]
  imul ebx, EQ_NUM_BANDS_ALIGNED_SIZE
  add esi, ebx

  // laden von A in EDX basierend auf Kanal position
  mov edx, [ebp+8]
  mov edx, [edx]
  add edx, ebx

  // laden des Buffer Zeigers in EDI anhand der aktuellen
  // Kanal Position von [EBP+20]
  mov edi, [esp]
  mov ebx, [ebp+20]
  // multiplizieren mit BitRate/8
  imul ebx, BITRATE
  // Zeigt auf Buffer+aktueller Channel
  add edi, ebx

  // Zhler EBX auf letzte Position des Buffers setzen
  mov ebx, [esp+8]
  imul ebx, [esp+4]
  add ebx, edi

@sampleLoop:
  // Aus Preformance Grnden, lieber mit 0.00000001 anstatt mit 0 multiplizieren
  // hat keine Auswirkung auf das Endergebniss
  cmp [edi], 0
  jne @sampleSkip
  movss xmm1, den
  jmp @sampleDo

@sampleSkip:
  movss xmm1, DWORD PTR [edi]

@sampleDo:
  mulss xmm1, xmm0

  // ESI FX
  // EDX FA

  // so := ( FA[c][n][5] * si +
  //         FA[c][n][0] * FX[c][n][0] +
  //         FA[c][n][1] * FX[c][n][1] -
  //         FA[c][n][2] * FX[c][n][2] -
  //         FA[c][n][3] * FX[c][n][3] )
  //       * FA[c][n][4];
  // FX[c][n][1] := FX[c][n][0];
  // FX[c][n][0] := si;
  // FX[c][n][3] := FX[c][n][2];
  // FX[c][n][2] := so;
  // si := so;

  mov ecx, EQ_NUM_BANDS
  shl ecx, 5
  add ecx, -32

@bandLoop:
  // EQ START

  movaps xmm3, DQWORD PTR [edx+ecx]
  movaps xmm5, DQWORD PTR [esi+ecx]
  mulps xmm3, xmm5
  movss xmm2, DWORD PTR [edx+ecx+20]
  mulss xmm2, xmm1
  xorps xmm3, xmm7
  movss [esi+ecx+4], xmm5

  db $f2,$0f,$7c,$db   // HADDPS xmm3, xmm3
  db $f2,$0f,$7c,$db   // HADDPS xmm3, xmm3

  addss xmm2, xmm3
  movss [esi+ecx], xmm1
  mulss xmm2, DWORD PTR [edx+ecx+16]
  mov eax, [esi+ecx+8]
  mov [esi+ecx+12], eax
  movss [esi+ecx+8], xmm2
  movss xmm1, xmm2

  // EQ ENDE

  add ecx, -32
  cmp ecx, -32
  jb @bandLoop

  // Range Check
  minss xmm1, MA
  maxss xmm1, MI

  // zurckschreiben
  movss [edi] ,xmm1

  add edi, [esp+8]
  cmp edi, ebx
  jne @sampleLoop

  cmp [ebp+20], 0
  jne @startLoop

  add esp, 12

  pop edi
  pop esi
  pop ebx
end;

{$IFDEF MEASURE}
function GetCPUCycles : Int64;
asm
  RDTSC
end;
{$ENDIF}

procedure TDCParamEQ.Process(Buffer: Pointer; Size, Samplerate: Integer; Bits: integer; Channels: integer; Float: Boolean; ChannelMapping: Cardinal);
{$IFDEF MEASURE}
var
  c1, c2: Int64;
{$ENDIF}
begin
  if not FEnabled then Exit;

  if FNeedCoeff or (FSampleRate <> Samplerate) or (FChannels <> Channels) or (FChannelMapping <> ChannelMapping) then
  begin
    FNeedCoeff := False;
    FSampleRate := Samplerate;
    FChannels := Channels;
    FChannelMapping := ChannelMapping;
    SetupChannels;
    CalcCoefficients;
    Flush;
  end;

  if (FNeedAmp) then
  begin
    CalcAmplitude;
    FNeedAmp := False;
  end;

{$IFDEF MEASURE}
  c1 := GetCPUCycles;
{$ENDIF}

  if hasSSE3 then
  begin
    case Bits of
      8:  Process8BitSSE3(Buffer, Size div Channels div (Bits div 8), Channels, @FChannelPreAmp, @FX, @FA);
      16: Process16BitSSE3(Buffer, Size div Channels div (Bits div 8), Channels, @FChannelPreAmp, @FX, @FA);
      24: Process24BitSSE3(Buffer, Size div Channels div (Bits div 8), Channels, @FChannelPreAmp, @FX, @FA);
      32: if Float
            then Process32BitFloatSSE3(Buffer, Size div Channels div (Bits div 8), Channels, @FChannelPreAmp, @FX, @FA)
            else Process32BitIntegerSSE3(Buffer, Size div Channels div (Bits div 8), Channels, @FChannelPreAmp, @FX, @FA);
    end;
  end else
  if hasSSE then
  begin
    case Bits of
      8:  Process8BitSSE(Buffer, Size div Channels div (Bits div 8), Channels, @FChannelPreAmp, @FX, @FA);
      16: Process16BitSSE(Buffer, Size div Channels div (Bits div 8), Channels, @FChannelPreAmp, @FX, @FA);
      24: Process24BitSSE(Buffer, Size div Channels div (Bits div 8), Channels, @FChannelPreAmp, @FX, @FA);
      32: if Float
            then Process32BitFloatSSE(Buffer, Size div Channels div (Bits div 8), Channels, @FChannelPreAmp, @FX, @FA)
            else Process32BitIntegerSSE(Buffer, Size div Channels div (Bits div 8), Channels, @FChannelPreAmp, @FX, @FA);
    end;
  end else
  begin
    case Bits of
      8:  Process8Bit(Buffer, Size div Channels div (Bits div 8), Channels);
      16: Process16Bit(Buffer, Size div Channels div (Bits div 8), Channels);
      24: Process24Bit(Buffer, Size div Channels div (Bits div 8), Channels);
      32: if Float
            then Process32BitFloat(Buffer, Size div Channels div (Bits div 8), Channels)
            else Process32Bit(Buffer, Size div Channels div (Bits div 8), Channels)
    end;
  end;
{$IFDEF MEASURE}
  c2 := GetCPUCycles;
  c2 := c2 - c1;
  if hasSSE3
    then OutputDebugString(PChar('EQ Measure SSE3: ' + inttostr(c2) + ' Cycles for ' + inttostr(Size) + ' bytes Difference: ' + inttostr(c2 div Size)))
  else if hasSSE
    then OutputDebugString(PChar('EQ Measure SSE: ' + inttostr(c2) + ' Cycles for ' + inttostr(Size) + ' bytes Difference: ' + inttostr(c2 div Size)))
    else OutputDebugString(PChar('EQ Measure Delphi: ' + inttostr(c2) + ' Cycles for ' + inttostr(Size) + ' bytes Difference: ' + inttostr(c2 div Size)));
{$ENDIF}
end;

procedure TDCParamEQ.Reset;
begin
  FChannelMapping := 0;
  FChannels := 0;
  FSampleRate := 0;
end;

function TDCParamEQ.GetChannelMapping: Cardinal;
var
  i: Integer;
begin
  if (FChannelMapping > 0) then
  begin
    Result := FChannelMapping;
  end else
  begin
    Result := 0;
    for i := 0 to FChannels -1
      do Inc(Result, 1 shl i);
  end;
end;

end.
